/* This file is part of the Project Athena Zephyr Notification System.
 * It contains source for the ZSubscribeTo, ZUnsubscribeTo, and
 * ZCancelSubscriptions functions.
 *
 *	Created by:	Robert French
 *
 *	$Source: /afs/andrew.cmu.edu/system/cvs/src/zephyr/lib/ZSubs.c,v $
 *	$Author: shadow $
 *
 *	Copyright (c) 1987,1988 by the Massachusetts Institute of Technology.
 *	For copying and distribution information, see the file
 *	"mit-copyright.h". 
 */
/* $Header: /afs/andrew.cmu.edu/system/cvs/src/zephyr/lib/ZSubs.c,v 1.2 1998/11/06 19:55:22 shadow Exp $ */

#include <internal.h>

#ifndef lint
static const char rcsid_ZSubscriptions_c[] = "$Id: ZSubs.c,v 1.2 1998/11/06 19:55:22 shadow Exp $";
#endif

static Code_t Z_Subscriptions __P((register ZSubscription_t *sublist,
				   int nitems, unsigned int port,
				   char *opcode, int authit));
static Code_t subscr_sendoff __P((ZNotice_t *notice, char **lyst, int num,
				  int authit));

#ifdef CMU_ZCTL_PUNT
Code_t ZPunt(sublist, nitems, port)
    ZSubscription_t *sublist;
    int nitems;
    unsigned int port;
{
    return (Z_Subscriptions(sublist, nitems, port, "SUPPRESS", 1));
}
#endif

Code_t ZSubscribeTo(sublist, nitems, port)
    ZSubscription_t *sublist;
    int nitems;
    unsigned int port;
{
    return (Z_Subscriptions(sublist, nitems, port, CLIENT_SUBSCRIBE, 1));
}

Code_t ZSubscribeToSansDefaults(sublist, nitems, port)
    ZSubscription_t *sublist;
    int nitems;
    unsigned int port;
{
    return (Z_Subscriptions(sublist, nitems, port, CLIENT_SUBSCRIBE_NODEFS,
			    1));
}

Code_t ZUnsubscribeTo(sublist, nitems, port)
    ZSubscription_t *sublist;
    int nitems;
    unsigned int port;
{
    return (Z_Subscriptions(sublist, nitems, port, CLIENT_UNSUBSCRIBE, 1));
}

Code_t ZCancelSubscriptions(port)
    unsigned int port;
{
    return (Z_Subscriptions((ZSubscription_t *)0, 0, port,
			    CLIENT_CANCELSUB, 0));
}

/*
 * This routine must do its own fragmentation.  Subscriptions must
 * not be broken across packet boundaries, or else the server will
 * mis-interpret them.
 */

static Code_t
Z_Subscriptions(sublist, nitems, port, opcode, authit)
    register ZSubscription_t *sublist;
    int nitems;
    unsigned int port;
    char *opcode;
    int authit;
{
    register int i, j;
    int retval;
    ZNotice_t notice;
    char header[Z_MAXHEADERLEN];
    char **list;
    char *recip;
    int hdrlen;
    int size_avail = Z_MAXPKTLEN-Z_FRAGFUDGE; /* space avail for data,
						 adjusted below */
    int size, start, numok;
    char *cp;

    /* nitems = 0 means cancel all subscriptions; still need to allocate a */
    /* array for one item so we can cancel, however. */
  
    list = (char **)malloc((unsigned)((nitems==0)?1:nitems)*3*sizeof(char *));
    if (!list)
        return (ENOMEM);

    (void) memset((char *)&notice, 0, sizeof(notice));
    notice.z_kind = ACKED;
    notice.z_port = port;
    notice.z_class = ZEPHYR_CTL_CLASS;
    notice.z_class_inst = ZEPHYR_CTL_CLIENT;
    notice.z_opcode = opcode;
    notice.z_sender = 0;
    notice.z_recipient = "";
    notice.z_default_format = "";
    notice.z_message_len = 0;

    /* format the header to figure out how long it is */
    retval = Z_FormatHeader(&notice, header, sizeof(header), &hdrlen, ZAUTH);
    if (retval != ZERR_NONE && !authit)
	retval = Z_FormatHeader(&notice, header, sizeof(header),
				&hdrlen, ZNOAUTH);
    if (retval != ZERR_NONE) {
	free((char *)list);
	return(retval);
    }

    /* compute amount of room left */
    size_avail -= hdrlen;
    size = size_avail;

    /* assemble subs into an array of pointers */
    for (i=0;i<nitems;i++) {
	list[i*3] = sublist[i].zsub_class;
	list[i*3+1] = sublist[i].zsub_classinst;
        if (!sublist[i].zsub_recipient || sublist[i].zsub_recipient[0] 
            == '\0' || (sublist[i].zsub_recipient[0] == '*' 
                        && sublist[i].zsub_recipient[1] != '@')) 
          {
            list[i*3+2] = "";
          } 
        else if (sublist[i].zsub_recipient[0] == '*' && 
                 sublist[i].zsub_recipient[1] == '@') 
          {
            cp = (char *)ZExpandRealm(sublist[i].zsub_recipient + 2);
            if (!(list[i*3+2] = malloc((unsigned)(strlen(cp) + 2)))) {
              for (i--; i >=0; i--)
                if (list[i*3+2][0] == '@')
                  free(list[i*3+2]);
              free((char *)list);
              return(ENOMEM);
            }
            (void) sprintf(list[i*3+2], "@%s", cp);
          } 
        else if (sublist[i].zsub_recipient[0] == '@') {
          cp = (char *) ZExpandRealm(sublist[i].zsub_recipient + 1);
          if (!(list[i*3+2] = malloc((unsigned)(strlen(cp) + 2)))) {
            for (i--; i >=0; i--)
              if (list[i*3+2][0] == '@')
                free(list[i*3+2]);
            free((char *)list);
            return(ENOMEM);
          }
          (void) sprintf(list[i*3+2], "@%s", cp);
        } 
        else
          {
            list[i*3+2] = ZGetSender();
          }
    }

    start = -1;
    i = 0;
    numok = 0;
    if (!nitems) {
	/* there aren't really any, but we need to xmit anyway */
	retval = subscr_sendoff(&notice, list, 0, authit);
          for (i=0;i<nitems;i++)
            if (list[i*3+2][0] == '@')
              free(list[i*3+2]);
	free((char *)list);
	return(retval);
    }
    while(i < nitems) {
	if (start == -1) {
	    size = size_avail;
	    start = i;
	    numok = 0;
	}
	if ((j = strlen(list[i*3])
	     + strlen(list[i*3+1])
	     + strlen(list[i*3+2]) + 3) <= size) {
	    /* it will fit in this packet */
	    size -= j;
	    numok++;
	    i++;
	    continue;
	}
	if (!numok) {			/* a single subscription won't
					   fit into one packet */
	    free((char *)list);
	    return(ZERR_FIELDLEN);
	}
	retval = subscr_sendoff(&notice, &list[start*3], numok, authit);
	if (retval) {
	    free((char *)list);
	    return(retval);
	}
	start = -1;
    }
    if (numok)
	retval = subscr_sendoff(&notice, &list[start*3], numok, authit);
    free((char *)list);
    return(retval);
}

static Code_t
subscr_sendoff(notice, lyst, num, authit)
ZNotice_t *notice;
char **lyst;
int num;
int authit;
{
    register Code_t retval;
    ZNotice_t retnotice;

    retval = ZSendList(notice, lyst, num*3, ZAUTH);
    if (retval != ZERR_NONE && !authit)
	retval = ZSendList(notice, lyst, num*3, ZNOAUTH);
	
    if (retval != ZERR_NONE)
	return (retval);
    if ((retval = ZIfNotice(&retnotice, (struct sockaddr_in *)0, 
				ZCompareUIDPred, (char *)&notice->z_uid)) !=
	ZERR_NONE)
	return (retval);
    if (retnotice.z_kind == SERVNAK) {
	ZFreeNotice(&retnotice);
	return (ZERR_SERVNAK);
    }
    if (retnotice.z_kind != SERVACK) {
	ZFreeNotice(&retnotice);
	return (ZERR_INTERNAL);
    }
    ZFreeNotice(&retnotice);
    return (ZERR_NONE);
}
